#!/usr/bin/env python3

import appdirs
from pathlib import Path
from configurator import Config
from functools import lru_cache

@lru_cache
def load_config():
    config = Config()
    cwd = Path.cwd()
    dirs = [cwd, *cwd.parents]

    for dir in reversed(dirs):
        paths = [
                dir / '.config' / 'po4' / 'conf.toml',
                dir / '.po4rc',
        ]
        for path in paths:
            if path.exists():
                subconf = Config.from_path(path, parser='toml')
                config.merge(subconf)

        dir = dir.parent

    return config.data
    
    



#!/usr/bin/env python3

import pytest, toml, inspect
from pathlib import Path
from functools import lru_cache

TEST_DIR = Path(__file__).parent

def parametrize_via_toml(relpath):

    def decorator(f):
        # The path is relative to the file the caller is defined in.
        module = inspect.getmodule(f)
        test_path = Path(module.__file__)
        toml_path = test_path.parent / relpath

        all_raw_params = cached_load(toml_path)
        try: raw_params = all_raw_params[f.__name__]
        except KeyError: raise ParametrizeViaTomlError(f"no parameters found for {f.__name__!r}") from None
        raw_args = set.union(*(set(x) for x in raw_params)) - {'id'}

        # Make sure there aren't any missing/extra parameters:
        for params in raw_params:
            if missing := raw_args - set(params):
                missing_str = ', '.join(f"'{x}'" for x in missing)
                raise ValueError(f"{toml_path}: {f.__name__}: missing parameter(s) {missing_str}")

        args = list(raw_args)
        params = [
                pytest.param(*(x[k] for k in args), id=x.get('id', None))
                for x in raw_params
        ]
        return pytest.mark.parametrize(args, params)(f)

    return decorator

@lru_cache
def cached_load(path):
    return toml.load(path)

class Params:

    @classmethod
    def parametrize(cls, f):
        args = cls.args

        params = []
        for k, v in cls.__dict__.items():
            if k.startswith('params'):
                params.extend(v)

        # Could also check to make sure parameters make sense.

        return pytest.mark.parametrize(args, params)(f)

class ParametrizeViaTomlError(Exception):
    pass




import sys, os
import po4

## General

project = 'PhosphateDB'
copyright = '2020, Kale Kundert'
version = po4.__version__
release = po4.__version__

master_doc = 'index'
source_suffix = '.rst'
templates_path = ['_templates']
exclude_patterns = ['_build']
default_role = 'any'

## Extensions

extensions = [
        'autoclasstoc',
        'sphinx.ext.autodoc',
        'sphinx.ext.autosummary',
        'sphinx.ext.viewcode',
        'sphinx.ext.intersphinx',
        'sphinx_rtd_theme',
]
intersphinx_mapping = { #
}
autosummary_generate = True
autodoc_default_options = {
        'exclude-members': '__dict__,__weakref__,__module__',
}
html_theme = 'sphinx_rtd_theme'
html_static_path = ['static']
pygments_style = 'sphinx'




#!/usr/bin/env python3

import pandas as pd
import autosnapgene as snap
from pathlib import Path
from voluptuous import Required, All, Any, Coerce
from ..model import Database, Tag, Plasmid, Fragment, Oligo
from ..protocols import Protocol
from ..utils import *

schema = {
        'type': 'excel',
        Required('dir'): All(str, Coerce(Path)),
        'columns': {str: Any(
            'seq', 'protocol', 'name', 'alt_names', 'date', 'desc', 
            'length', 'conc', 
        )},
}
default_config = {
        'columns': {
            'Sequence': 'seq',
            'Construction': 'protocol',
            'Name': 'name',
            'Cross-refs': 'alt_names',
            'Date': 'date',
            'Description': 'desc',
            'Length': 'length',
            'Conc': 'conc',
        }
}

def load(config):
    root = config['dir']

    db_xlsx = {
            Plasmid:  root / 'plasmids.xlsx',
            Fragment: root / 'fragments.xlsx',
            Oligo:    root / 'oligos.xlsx',
    }
    seq_dirs = {
            Plasmid:  root / 'plasmids',
            Fragment: root / 'fragments',
            Oligo:    root / 'oligos',
    }

    for dir in seq_dirs.values():
        dir.mkdir(exist_ok=True, parents=True)

    db = Database(str(root))

    for cls, path in db_xlsx.items():
        df = pd.read_excel(path)
        df = df.set_index(df.index + 2)
        df = df.rename(columns=config['columns'])
        df = df.where(pd.notnull(df), None)

        for i, row in df.iterrows():
            tag = Tag(cls.tag_prefix, i)
            kwargs = dict(row)

            if not kwargs.get('seq'):
                kwargs['seq'] = _defer(_seq_from_tag, seq_dirs[cls], tag)
            if x := kwargs.get('alt_names'):
                kwargs['alt_names'] = x.split(',')
            if x := kwargs.get('protocol'):
                kwargs['protocol'] = _defer(Protocol.from_text, db, x)

            db[tag] = cls(**kwargs)

    return db

def _defer(f, *args, **kwargs):
    """
    Create a closure that call the given function with the given arguments.

    The point of this function is to avoid the surprising behavior that can 
    occur if you define a closure in a scope where variables are changing (e.g. 
    in a for-loop).  The confusing thing is that closures have access to the 
    scope they were defined in, but only as it exists when they are ultimately 
    called.  So if the scope changes between when the closure is defined and 
    when it's called, the closure will use the final value of any variables.

    This function serves to create a static local scope containing the 
    variables needed by the closure, which avoids the problem.

    More information: 
    https://stackoverflow.com/questions/10452770/python-lambdas-binding-to-local-values
    """
    return lambda: f(*args, **kwargs)

def _seq_from_tag(dir, tag):
    if path := _path_from_tag(dir, tag):
        return snap.parse(path).dna_sequence

def _path_from_tag(dir, tag):
    # Figure out the path to the actual file, allowing for several different 
    # naming conventions.
    for path in dir.iterdir():
        if re.fullmatch(f'({tag.type})?0*{tag.id}([_-].*)?.dna', path.name):
            return path




#!/usr/bin/env python3

import os
from contextlib import contextmanager
from pathlib import Path
from po4 import load_config

DUMMY_CONFIG = Path(__file__).parent / 'dummy_config'

def test_config():
    with cd(DUMMY_CONFIG):
        load_config.cache_clear()
        config = load_config()

        # Only check the values that are explicitly set by the test, because 
        # any other values could be affected by real configuration files 
        # present on the tester's system.

        assert config['use'] == 'db1'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'

    with cd(DUMMY_CONFIG / 'subdir'):
        load_config.cache_clear()
        config = load_config()

        assert config['use'] == 'db2'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'


@contextmanager
def cd(dir):
    try:
        prev_cwd = Path.cwd()
        os.chdir(dir)
        yield

    finally:
        os.chdir(prev_cwd)



#!/usr/bin/env python3

import pytest
from po4 import only_raise

class ParentError(Exception):
    pass

class ChildError(ParentError):
    pass

class UnrelatedError(Exception):
    pass

@pytest.mark.parametrize(
        'only_err, raise_err, catch_err', [
            (   ParentError,    ParentError,    ParentError),
            (   ParentError,     ChildError,     ChildError),
            (   ParentError, UnrelatedError,    ParentError),

            (    ChildError,    ParentError,     ChildError),
            (    ChildError,     ChildError,     ChildError),
            (    ChildError, UnrelatedError,     ChildError),

            (UnrelatedError,    ParentError, UnrelatedError),
            (UnrelatedError,     ChildError, UnrelatedError),
            (UnrelatedError, UnrelatedError, UnrelatedError),
])
def test_only_raise(only_err, raise_err, catch_err):

    @only_raise(only_err)
    def f(err):
        raise err

    with pytest.raises(catch_err):
        f(raise_err)




#!/usr/bin/env python3

import pytest
from po4 import Tag, ParseError
from po4.model import Construct, Plasmid, Fragment, Oligo
from po4.utils import *
from utils import Params, parametrize_via_toml
from test_model import DummyConstruct

class PrefixParams(Params):
    args = 'args, expected'
    params = [
            ([Construct, Plasmid, Fragment, Oligo], 'dfop'),
            ([Construct, Plasmid, Fragment,      ], 'dfop'),
            ([Construct, Plasmid,           Oligo], 'dfop'),
            ([Construct, Plasmid,                ], 'dfop'),
            ([Construct,          Fragment, Oligo], 'dfop'),
            ([Construct,          Fragment,      ], 'dfop'),
            ([Construct,                    Oligo], 'dfop'),
            ([Construct,                         ], 'dfop'),
            ([           Plasmid, Fragment, Oligo], 'fop'),
            ([           Plasmid, Fragment,      ], 'fp'),
            ([           Plasmid,           Oligo], 'op'),
            ([           Plasmid,                ], 'p'),
            ([                    Fragment, Oligo], 'fo'),
            ([                    Fragment,      ], 'f'),
            ([                              Oligo], 'o'),
            ([                                   ], 'dfop'),
    ]

@parametrize_via_toml('test_utils.toml')
def test_normalize_seq(raw_seq, expected):
    assert normalize_seq(raw_seq) == expected

@PrefixParams.parametrize
def test_get_tag_prefixes(args, expected):
    assert get_tag_prefixes(*args) == set(expected)

@PrefixParams.parametrize
def test_get_tag_pattern(args, expected):
    assert get_tag_pattern(*args) == fr'[{expected}]\d+'

@parametrize_via_toml('test_utils.toml')
def test_parse_tag(tag_str, expected):
    assert parse_tag(tag_str) == Tag(**expected)

@parametrize_via_toml('test_utils.toml')
def test_parse_tag_err(tag_str, error):
    with pytest.raises(ParseError, match=error):
        parse_tag(tag_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_params(params_str, expected):
    assert parse_params(params_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_params_err(params_str, error):
    with pytest.raises(ParseError, match=error):
        parse_params(params_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_param(params, key, pattern, default, expected):
    if isinstance(expected, list):
        expected = tuple(expected)
    assert parse_param(params, key, pattern, default) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_param_err(params, key, pattern, error):
    with pytest.raises(ParseError, match=error):
        parse_param(params, key, pattern)

@parametrize_via_toml('test_utils.toml')
def test_parse_bool(bool_str, expected):
    assert parse_bool(bool_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_bool_err(bool_str, error):
    with pytest.raises(ParseError, match=error):
        parse_bool(bool_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_time_s(time_str, expected):
    assert parse_time_s(time_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_time_s_err(time_str, error):
    with pytest.raises(ParseError, match=error):
        parse_time_s(time_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_temp_C(temp_str, expected):
    assert parse_temp_C(temp_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_temp_C_err(temp_str, error):
    with pytest.raises(ParseError, match=error):
        parse_temp_C(temp_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_volume_uL(vol_str, expected):
    assert parse_volume_uL(vol_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_volume_uL_err(vol_str, error):
    with pytest.raises(ParseError, match=error):
        parse_volume_uL(vol_str)

@parametrize_via_toml('test_utils.toml')
def test_parse_conc_nM(conc_str, mw, expected_nM, expected_ng_uL):
    assert parse_conc_nM(conc_str, mw)    == pytest.approx(expected_nM)
    assert parse_conc_uM(conc_str, mw)    == pytest.approx(expected_nM / 1000)
    assert parse_conc_ng_uL(conc_str, mw) == pytest.approx(expected_ng_uL)

@parametrize_via_toml('test_utils.toml')
def test_parse_conc_err(conc_str, mw, error):
    with pytest.raises(ParseError, match=error):
        parse_conc_nM(conc_str, mw)
    with pytest.raises(ParseError, match=error):
        parse_conc_ng_uL(conc_str, mw)
@parametrize_via_toml('test_utils.toml')
def test_parse_size_bp(size_str, expected):
    assert parse_size_bp(size_str) == expected

@parametrize_via_toml('test_utils.toml')
def test_parse_size_bp_err(size_str, error):
    with pytest.raises(ParseError, match=error):
        parse_size_bp(size_str)




#!/usr/bin/env python3

import autoprop
from more_itertools import one
from dataclasses import dataclass
from collections import namedtuple
from configurator import Config
from voluptuous import Schema, Invalid
from pkg_resources import iter_entry_points
from Bio.SeqUtils import molecular_weight, MeltingTemp
from .config import load_config
from .errors import LoadError, QueryError, CheckError, only_raise
from .utils import *

class Database:

    def __init__(self, name=None):
        self.name = name
        self._constructs = {}

    def __iter__(self):
        yield from self._constructs

    def __len__(self):
        return len(self._constructs)

    def __getitem__(self, tag):
        tag = parse_tag(tag)
        try:
            return self._constructs[tag]
        except KeyError:
            raise QueryError(f"not found in database", culprit=tag) from None

    def __setitem__(self, tag, construct):
        tag = parse_tag(tag)
        if tag in self._constructs:
            raise LoadError(f"already in database, cannot be replaced", culprit=tag)
        if tag.type != construct.tag_prefix:
            raise LoadError(f"{construct} cannot have tag '{tag}': expected {construct.tag_prefix!r} prefix")

        self._constructs[tag] = construct
        construct._db = self
        construct._tag = tag

    def __delitem__(self, tag):
        construct = self._constructs.pop(parse_tag(tag))
        construct._db = None
        construct._tag = None

    def __contains__(self, construct):
        return construct._tag in self._constructs

    def keys(self):
        return self._constructs.keys()

    def values(self):
        return self._constructs.values()

    def items(self):
        return self._constructs.items()


@dataclass(frozen=True)
class Tag:
    type: str
    id: int

    def __str__(self):
        return f'{self.type}{self.id}'

@autoprop
class Construct:
    tag_prefix = None

    def __init__(self, **kwargs):
        self._db = None
        self._tag = None

        self._seq = kwargs.get('seq')
        self._name = kwargs.get('name')
        self._alt_names = kwargs.get('alt_names', [])
        self._date = kwargs.get('date')
        self._desc = kwargs.get('desc')
        self._length = kwargs.get('length')
        self._conc_str = kwargs.get('conc')
        self._protocol = kwargs.get('protocol')

    def check(self):
        self._check_seq()

    def _check_seq(self):
        from Bio import pairwise2
        from Bio.pairwise2 import format_alignment

        try:
            primary_seq = self.seq
            protocol_seq = self.protocol.product_seq
        except (QueryError, NotImplementedError):
            pass
        else:
            if primary_seq != protocol_seq:
                alignments = pairwise2.align.globalxx(primary_seq, protocol_seq)
                message = f"sequence doesn't match protocol\n" + format_alignment(*alignments[0])
                raise CheckError(message, culprit=self._tag)

    def get_db(self):
        if not self._db:
            raise QueryError("not attached to a database", culprit=self._tag)
        return self._db

    def get_tag(self):
        if not self._tag:
            raise QueryError("not attached to a database", culprit=self._tag)
        return self._tag

    def get_seq(self):
        # Allow the retrieval of the sequence to be deferred, e.g. so unused 
        # sequences never have to be read from disc.
        if callable(self._seq):
            self._seq = self._seq()

        # If we have instructions for how to make the construct, try getting
        # the sequence from that.
        if not self._seq and self._protocol:
            self._seq = self.protocol.product_seq

        if not self._seq:
            raise QueryError("no sequence specified", culprit=self._tag)

        # Always return a `Bio.Seq` object.
        if isinstance(self._seq, str):
            self._seq = DnaSeq(self._seq)

        return self._seq

    def get_length(self):
        if self._length:
            return self._length
        else:
            return len(self.seq)

    def get_name(self):
        return self._name

    def get_alt_names(self):
        return self._alt_names

    def get_date(self):
        return self._date

    def get_desc(self):
        return self._desc

    def get_protocol(self):
        if not self._protocol:
            raise QueryError("no protocol specified", culprit=self._tag)

        # Allow `self._protocol` to be a callable that returns a protocol, so 
        # that the protocol doesn't have to be loaded until it is actually 
        # needed.  This means lets the database load faster, and avoids 
        # generating errors relating to protocols that the user doesn't 
        # actively care about.
        if callable(self._protocol):
            self._protocol = self._protocol()

        return self._protocol

    def get_mw(self):
        from Bio.SeqUtils import molecular_weight
        mw = molecular_weight(
                seq=self.seq,
                double_stranded=self.is_double_stranded,
                circular=self.is_circular,
        )

        # For some reason Biopython just assumes 5' phosphorylation, so we need 
        # to correct for that here.
        hpo3 = 1.008 + 30.974 + 3*15.999
        if not self.is_phosphorylated:
            strands = 2 if self.is_double_stranded else 1
            ends = 0 if self.is_circular else strands
            mw -= hpo3 * ends

        return mw

    def get_conc_str(self):
        if not self._conc_str:
            raise QueryError("no concentration specified", culprit=self._tag)
        return self._conc_str

    def get_conc_nM(self):
        return parse_conc_nM(self.conc_str, self.mw)

    def get_conc_ng_uL(self):
        return parse_conc_ng_uL(self.conc_str, self.mw)

    @property
    def is_double_stranded(self):
        raise NotImplementedError

    @property
    def is_single_stranded(self):
        return not self.is_double_stranded

    @property
    def is_circular(self):
        raise NotImplementedError

    @property
    def is_linear(self):
        return not self.is_circular

    def is_phosphorylated(self):
        raise NotImplementedError


@autoprop
class Plasmid(Construct):
    tag_prefix = 'p'

    @property
    def is_double_stranded(self):
        return True

    @property
    def is_circular(self):
        return True

@autoprop
class Fragment(Construct):
    tag_prefix = 'f'

    @property
    def is_double_stranded(self):
        return self.protocol.is_product_double_stranded

    @property
    def is_circular(self):
        return False

    @property
    def is_phosphorylated(self):
        return self.protocol.is_product_phosphorylated


@autoprop
class Oligo(Construct):
    tag_prefix = 'o'

    def get_melting_temp(self):
        from Bio.SeqUtils import MeltingTemp

        # If the TM is encoded in the oligo name, use that.
        if m := re.search(r'[-_ ](TM|Tm|tm)=?(\d+)', self.name):
            return float(m.group(2))

        # Otherwise, calculate a Tm using the Wallace rule.  This isn't a 
        # particularly accurate method, but I chose it because it agrees most 
        # closely with NEB's Tm calculator, which is what I've been using for 
        # everything.
        else:
            return MeltingTemp.Tm_Wallace(self.seq)

    def get_tm(self):
        return self.melting_temp

    @property
    def is_double_stranded(self):
        return False

    @property
    def is_circular(self):
        return False

    @property
    def is_phosphorylated(self):
        # Right now, PO4 has no support for modified oligos at all.  
        return False

@only_raise(LoadError)
def load_db(use=None, config=None):
    if not config:
        # Can't test this line, because it reads the real configuration files 
        # on the tester's machine, and so cannot be made to produce consistent 
        # results.
        config = load_config()  # pragma: no cover

    if not use:
        try:
            use = config['use']
        except KeyError as err:
            raise LoadError("no database specified.") from None

    try:
        config_use = config['database'][use]
    except KeyError as err:
        raise LoadError(f"unknown database {use!r}") from None

    try:
        type_use = config_use['type']
    except KeyError as err:
        raise LoadError(f"no 'type' specified for database {use!r}") from None

    plugin = one(
            iter_entry_points('po4.databases', type_use),
            too_short=LoadError(f"no {type_use!r} database plugin found."),
            too_long=LoadError(f"multiple {type_use!r} database plugins found."),
    )
    plugin = plugin.load()

    if defaults := getattr(plugin, 'default_config'):
        config_use = (Config(defaults) + config_use).data

    if hasattr(plugin, 'schema'):
        schema = Schema(plugin.schema)
        config_use = schema(config_use)

    return plugin.load(config_use)




#!/usr/bin/env python3

import pytest
import shlex
from po4 import Database, Tag, Protocol, Plasmid, Fragment, Oligo
from po4.stepwise import Make
from test_model import DummyConstruct
from utils import parametrize_via_toml

@parametrize_via_toml('test_make.toml')
def test_make(disable_capture, protocols, options, expected):
    db = Database()
    t7 = 'TAATACGACTCACTATA'

    # Create some dummy sequences for the protocols to use.
    db['p1'] = Plasmid(seq='GATTACA')
    db['p2'] = Plasmid(seq='GATTACA')
    db['f1'] = Fragment(seq='GATTACA')
    db['f2'] = Fragment(seq='GATTACA', conc='100ng/µL')
    db['f3'] = Fragment(seq='GATTACA', conc='75nM')
    db['f4'] = Fragment(seq='GATTACA')
    db['f5'] = Fragment(seq='GATTACA', conc='100ng/µL')
    db['f6'] = Fragment(seq='GATTACA', conc='75nM')
    db['f7'] = Fragment(seq=t7+300*'G')
    db['f8'] = Fragment(seq=t7+301*'G')
    db['o1'] = Oligo(seq='GATTACA')
    db['o2'] = Oligo(seq='GATTACA')
    db['o3'] = Oligo(seq='GATTACA')
    db['o4'] = Oligo(seq='GATTACA')

    tags = []
    for i, protocol_str in enumerate(protocols, 1):
        tag = Tag('d', i); tags.append(tag)
        protocol = Protocol.from_text(db, protocol_str)
        db[tag] = DummyConstruct(protocol=protocol)

    make = Make(db, tags, options)
    with disable_capture:
        p = make.protocol
    
    for cmd_i, expected_i in zip(p.commands, expected):
        assert cmd_i == expected_i

# Copied from stepwise
@pytest.fixture
def disable_capture(pytestconfig):
    # Equivalent to `pytest -s`, but temporary.
    # This is necessary because even `capfd.disabled()` leaves stdin in a state 
    # that somehow interferes with the redirection we're trying to do.

    class suspend_guard:

        def __init__(self):
            self.capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')

        def __enter__(self):
            self.capmanager.suspend_global_capture(in_=True)
            pass

        def __exit__(self, _1, _2, _3):
            self.capmanager.resume_global_capture()

    yield suspend_guard()





#!/usr/bin/env python3

import pytest
from po4 import Database, Plasmid, Fragment, Oligo
from po4.protocols import *
from utils import *

@autoprop
class DummyProtocol(Protocol):
    name = 'DUMMY'

    def __init__(self, seq, db=None):
        super().__init__(db)
        self._seq = seq

    @classmethod
    def from_params(cls, db, params):
        return cls(params['seq'], db)

    def get_product_seq(self):
        return self._seq

class DummyProtocolSubclass(DummyProtocol):
    name = 'DUMMY_SUBCLS'

def test_protocol_from_text():
    db = Database()
    dummy1 = Protocol.from_text(db, 'DUMMY: seq=GATTACA')

    assert isinstance(dummy1, DummyProtocol)
    assert dummy1.db is db
    assert dummy1.product_seq == 'GATTACA'

    dummy2 = Protocol.from_text(db, 'DUMMY_SUBCLS: seq=TGTAATC')
    assert isinstance(dummy2, DummyProtocolSubclass)
    assert dummy2.db is db
    assert dummy2.product_seq == 'TGTAATC'

    with pytest.raises(ParseError, match="UNDEFINED: unknown protocol type"):
        Protocol.from_text(db, 'UNDEFINED: seq=TGTAATC')


def test_pcr_from_params():
    db = Database()
    pcr = PcrProtocol.from_params(db, dict(
            template='f1',
            primers='o1,o2',
            Ta='60°C',
            tx='30s',
            volume='50 µL',
    ))
    assert pcr.template_tag == 'f1'
    assert pcr.primer_tags == ('o1', 'o2')
    assert pcr.annealing_temp_C == 60
    assert pcr.extension_time_s == 30
    assert pcr.volume_uL == 50

@parametrize_via_toml('test_protocols.toml')
def test_pcr_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        PcrProtocol.from_params(None, params)

def test_pcr_template():
    db = Database()
    db['p1'] = Plasmid(seq='AAAA')
    db['f1'] = Fragment(seq='CCCC')

    pcr1 = PcrProtocol(db, 'p1', ('o1', 'o2'))
    assert pcr1.template_tag == 'p1'
    assert pcr1.template_seq == 'AAAA'

    pcr2 = PcrProtocol(db, 'f1', ('o1', 'o2'))
    assert pcr2.template_tag == 'f1'
    assert pcr2.template_seq == 'CCCC'

def test_pcr_primers():
    db = Database()
    db['o1'] = Oligo(seq='AAAA')
    db['o2'] = Oligo(seq='CCCC')

    pcr = PcrProtocol(db, 'f1', ('o1', 'o2'))
    assert pcr.primer_seqs == ['AAAA', 'CCCC']

@parametrize_via_toml('test_protocols.toml')
def test_pcr_product(primer_fwd, primer_rev, template, product):
    db = Database()
    db['p1'] = Plasmid(seq=template)
    db['o1'] = Oligo(seq=primer_fwd)
    db['o2'] = Oligo(seq=primer_rev)

    pcr = PcrProtocol(db, 'p1', ('o1', 'o2'))
    assert pcr.product_seq == product.upper()
    assert pcr.product_len == len(product)

@parametrize_via_toml('test_protocols.toml')
def test_pcr_product_err(primer_fwd, primer_rev, template, error):
    db = Database()
    db['f1'] = Fragment(seq=template)
    db['o1'] = Oligo(seq=primer_fwd)
    db['o2'] = Oligo(seq=primer_rev)

    pcr = PcrProtocol(db, 'f1', ('o1', 'o2'))

    with pytest.raises(QueryError, match=error):
        pcr.product_seq

def test_pcr_annealing_temp():
    db = Database()
    db['o1'] = Oligo(name='O1_TM60')
    db['o2'] = Oligo(name='O2_TM62')

    pcr1 = PcrProtocol(db, 'f1', ('o1', 'o2'))
    assert pcr1.annealing_temp_C == 61  # Lowest primer Tm +1°C

    pcr2 = PcrProtocol(db, 'f1', ('o1', 'o2'), Ta=60)
    assert pcr2.annealing_temp_C == 60

@parametrize_via_toml('test_protocols.toml')
def test_pcr_extension_time(product_len, extension_time_s):
    db = Database()
    db['o1'] = o1 = Oligo(seq='tataacaggctgctgagacc')  # SR001
    db['o2'] = o2 = Oligo(seq='tctaggactatcaccggagg')  # SR002

    n = product_len - len(o1.seq) - len(o2.seq)
    db['f1'] = Fragment(seq=f'{o1.seq}{n * "A"}{o2.seq.reverse_complement()}')

    pcr = PcrProtocol(db, 'f1', ('o1', 'o2'))
    assert pcr.product_len == product_len
    assert pcr.extension_time_s == pytest.approx(extension_time_s)

def test_pcr_volume():
    pcr1 = PcrProtocol(None, None, None)
    assert pcr1.volume_uL == None

    pcr2 = PcrProtocol(None, None, None, volume=10)
    assert pcr2.volume_uL == 10


def test_digest_from_params():
    db = Database()
    digest = DigestProtocol.from_params(db, dict(
            template='p1',
            enzymes='BsaI',
            size='1kb',
    ))
    assert digest.template_tag == 'p1'
    assert digest.enzymes == ['BsaI']
    assert digest._product_size == 1000

@parametrize_via_toml('test_protocols.toml')
def test_digest_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        DigestProtocol.from_params(None, params)

@parametrize_via_toml('test_protocols.toml')
def test_digest_product(seq, enzymes, circular, size, products, product):
    Construct = Plasmid if circular else Fragment
    tag = f'{Construct.tag_prefix}1'

    db = Database()
    db[tag] = Construct(seq=seq)

    digest = DigestProtocol(db, tag, enzymes, size or None)
    assert digest.product_seqs == [DnaSeq(x) for x in products]
    assert digest.product_seq == DnaSeq(product)

@parametrize_via_toml('test_protocols.toml')
def test_digest_product_err(seq, enzymes, error):
    db = Database()
    db['p1'] = Plasmid(seq=seq)

    digest = DigestProtocol(db, 'p1', enzymes)

    with pytest.raises(QueryError, match=error):
        digest.product_seq


def test_anneal_from_params():
    db = Database()
    db['o1'] = Oligo(seq='GATTACA')
    db['o2'] = Oligo(seq='TGTAATC')

    anneal = AnnealProtocol.from_params(db, dict(
            oligos='o1,o2',
            volume='10µL',
            conc='50µM',
            stock='200µM',
    ))
    assert anneal.oligo_tags == ('o1', 'o2')
    assert anneal.volume_uL == 10
    assert anneal.conc_uM == 50
    assert anneal.stock_uM == 200

@parametrize_via_toml('test_protocols.toml')
def test_anneal_from_params_err(params, error):
    db = Database()
    db['o1'] = Oligo(seq='GATTACA')
    db['o2'] = Oligo(seq='TGTAATC')

    with pytest.raises(ParseError, match=error):
        AnnealProtocol.from_params(db, params)


def test_gg_from_params():
    db = Database()
    gg = GoldenGateProtocol.from_params(db, dict(
            bb='p1',
            ins='f1,f2',
            enzyme='BbsI',
            volume='10µL',
    ))
    assert gg.backbone_tag == 'p1'
    assert gg.insert_tags == ['f1', 'f2']
    assert gg.enzyme == 'BbsI'
    assert gg.volume_uL == 10

@parametrize_via_toml('test_protocols.toml')
def test_gg_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        GoldenGateProtocol.from_params(None, params)

def test_gg_bb_ins():
    db = Database()
    db['p1'] = Plasmid(seq='AAAA')
    db['f1'] = Fragment(seq='CCCC')
    db['f2'] = Fragment(seq='GGGG')

    gg = GoldenGateProtocol(db, 'p1', ['f1', 'f2'])
    assert gg.backbone_seq == 'AAAA'
    assert gg.insert_seqs == ['CCCC', 'GGGG']


def test_gib_from_params():
    db = Database()
    gib = GibsonProtocol.from_params(db, dict(
            bb='p1',
            ins='f1,f2',
            volume='10µL',
    ))
    assert gib.backbone_tag == 'p1'
    assert gib.insert_tags == ['f1', 'f2']
    assert gib.volume_uL == 10


def test_ligate_from_params():
    db = Database()
    lig = LigateProtocol.from_params(db, dict(
            bb='f1',
            ins='f2,f3',
            volume='10µL',
            kinase='y',
    ))
    assert lig.backbone_tag == 'f1'
    assert lig.insert_tags == ['f2', 'f3']
    assert lig.volume_uL == 10
    assert lig.use_kinase == True

@parametrize_via_toml('test_protocols.toml')
def test_ligate_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        LigateProtocol.from_params(None, params)


def test_ivt_from_params():
    db = Database()
    ivt = IvtProtocol.from_params(db, dict(
            template='f1',
    ))
    assert ivt.template_tag == 'f1'
    assert ivt.is_product_double_stranded == False

@parametrize_via_toml('test_protocols.toml')
def test_ivt_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        IvtProtocol.from_params(None, params)

@parametrize_via_toml('test_protocols.toml')
def test_ivt_product(template, expected):
    from Bio.Seq import Seq
    from Bio.Alphabet import RNAAlphabet

    db = Database()
    db['f1'] = Fragment(seq=template)

    ivt = IvtProtocol(db, 'f1')
    assert ivt.product_seq == Seq(expected, RNAAlphabet())

@parametrize_via_toml('test_protocols.toml')
def test_ivt_product_err(template, error):
    db = Database()
    db['f1'] = Fragment(seq=template)

    ivt = IvtProtocol(db, 'f1')

    with pytest.raises(QueryError, match=error):
        ivt.product_seq


def test_idt_from_params():
    db = Database()
    idt = IdtProtocol.from_params(db, dict(
            seq='GATTACA',
    ))
    assert idt.product_seq == 'GATTACA'

@parametrize_via_toml('test_protocols.toml')
def test_idt_from_params_err(params, error):
    with pytest.raises(ParseError, match=error):
        IdtProtocol.from_params(None, params)




#!/usr/bin/env python3

import pytest
import datetime
import autoprop
from po4 import QueryError, DnaSeq, Protocol
from po4.model import *
from utils import *
from test_protocols import DummyProtocol

class DummyConstruct(Construct):
    tag_prefix = 'd'

    @property
    def is_double_stranded(self):
        return True

    @property
    def is_circular(self):
        return False

def test_database_getitem_setitem():
    db = Database()

    d0 = DummyConstruct()
    with pytest.raises(QueryError, match='not attached'): d0.db
    with pytest.raises(QueryError, match='not attached'): d0.tag
    with pytest.raises(QueryError, match="d0: not found in database"): db['d0']

    db['d1'] = d1 = DummyConstruct()
    db['d02'] = d2 = DummyConstruct()
    db['d', 3] = d3 = DummyConstruct()
    db[Tag('d', 4)] = d4 = DummyConstruct()

    with pytest.raises(LoadError, match="d1: already in database, cannot be replaced"):
        db['d1'] = DummyConstruct()
    with pytest.raises(LoadError, match="<.*DummyConstruct.*> cannot have tag 'p1': expected 'd' prefix"):
        db['p1'] = DummyConstruct()

    assert d1.db is db
    assert d2.db is db
    assert d3.db is db
    assert d4.db is db

    assert d1.tag == Tag('d', 1)
    assert d2.tag == Tag('d', 2)
    assert d3.tag == Tag('d', 3)
    assert d4.tag == Tag('d', 4)

    assert db['d1'] is d1
    assert db['d01'] is d1
    assert db['d', 1] is d1
    assert db[Tag('d', 1)] is d1

    assert db['d2'] is d2
    assert db['d02'] is d2
    assert db['d', 2] is d2
    assert db[Tag('d', 2)] is d2

    assert db['d3'] is d3
    assert db['d03'] is d3
    assert db['d', 3] is d3
    assert db[Tag('d', 3)] is d3

    assert db['d4'] is d4
    assert db['d04'] is d4
    assert db['d', 4] is d4
    assert db[Tag('d', 4)] is d4

def test_database_delitem():
    db = Database()
    db['d1'] = d1 = DummyConstruct()

    assert d1 in db
    assert d1.db is db
    assert d1.tag == Tag('d', 1)

    del db['d1']

    assert d1 not in db
    with pytest.raises(QueryError, match='not attached'): d1.db
    with pytest.raises(QueryError, match='not attached'): d1.tag

def test_database_contains():
    db = Database()
    d1 = DummyConstruct()
    assert d1 not in db

    db['d1'] = d1
    assert d1 in db

def test_database_iter():
    db = Database()
    db['d1'] = d1 = DummyConstruct()
    db['d2'] = d2 = DummyConstruct()

    values = {d1, d2}
    keys = {x.tag for x in values}
    items = {(x.tag, x) for x in values}

    assert set(db) == keys
    assert set(db.keys()) == keys
    assert set(db.values()) == values
    assert set(db.items()) == items

def test_database_len():
    db = Database()
    assert len(db) == 0

    db['d1'] = DummyConstruct()
    assert len(db) == 1

    db['d2'] = DummyConstruct()
    assert len(db) == 2


def test_tag():
    tag = Tag('p', 1)

    assert tag.type == 'p'
    assert tag.id == 1
    assert str(tag) == 'p1'


@parametrize_via_toml('test_model.toml')
def test_construct_seq(given_seq, protocol_seq, expected):
    kwargs = {}

    if given_seq != False:
        seq = given_seq['seq']
        kwargs['seq'] = (lambda: seq) if given_seq['defer'] else seq
    if protocol_seq != False:
        kwargs['protocol'] = DummyProtocol(protocol_seq)

    d1 = DummyConstruct(**kwargs)
    assert d1.seq == DnaSeq(expected)

def test_construct_seq_err():
    d1 = DummyConstruct()
    with pytest.raises(QueryError, match='no sequence specified'):
        d1.seq

@parametrize_via_toml('test_model.toml')
def test_construct_length(columns, expected):
    d1 = DummyConstruct(**columns)
    assert d1.length == expected

def test_construct_length_err():
    d1 = DummyConstruct()
    with pytest.raises(QueryError, match='no sequence specified'):
        d1.length

@parametrize_via_toml('test_model.toml')
def test_construct_name(columns, expected):
    d1 = DummyConstruct(**columns)
    assert d1.name == (expected or None)

@parametrize_via_toml('test_model.toml')
def test_construct_date(columns, expected):
    d1 = DummyConstruct(**columns)
    assert d1.date == (expected or None)

@parametrize_via_toml('test_model.toml')
def test_construct_desc(columns, expected):
    d1 = DummyConstruct(**columns)
    assert d1.desc == (expected or None)

def test_construct_protocol():
    d1 = DummyConstruct()

    with pytest.raises(QueryError, match="no protocol"):
        d1.protocol

    protocol = DummyProtocol('ATCG')
    d2 = DummyConstruct(protocol=protocol)
    d3 = DummyConstruct(protocol=lambda: protocol)

    assert d2.protocol is protocol
    assert d3.protocol is protocol

def test_construct_conc():
    d1 = DummyConstruct()

    with pytest.raises(QueryError, match="no concentration"):
        d1.conc_str

    d2 = DummyConstruct(seq='ATCG', conc='10 ng/µL')

    assert d2.conc_str == '10 ng/µL'
    assert d2.conc_ng_uL == pytest.approx(10)

    d3 = DummyConstruct(seq='ATCG', conc='10 nM')

    assert d3.conc_str == '10 nM'
    assert d3.conc_nM == pytest.approx(10)


def test_plasmid_mw():
    # http://molbiotools.com/dnacalculator.html
    p1 = Plasmid(seq='ATCG')

    assert p1.is_circular == (not p1.is_linear) == True
    assert p1.is_double_stranded == (not p1.is_single_stranded) == True
    assert p1.mw == pytest.approx(2471.58, rel=1e-3)

def test_fragment_mw():
    # 5'-phosphorylation assumed.
    # http://molbiotools.com/dnacalculator.html
    f1 = Fragment(protocol=DummyProtocol('ATCG'))

    assert f1.is_circular == (not f1.is_linear) == False
    assert f1.is_double_stranded == (not f1.is_single_stranded) == True
    assert f1.mw == pytest.approx(2507.61, rel=1e-3)

def test_oligo_mw():
    # 5'-OH assumed.
    # http://molbiotools.com/dnacalculator.html
    o1 = Oligo(seq='ATCG')

    assert o1.is_circular == (not o1.is_linear) == False
    assert o1.is_double_stranded == (not o1.is_single_stranded) == False
    assert o1.mw == pytest.approx(1173.82, rel=1e-3)

@parametrize_via_toml('test_model.toml')
def test_oligo_tm(seq, name, expected):
    o1 = Oligo(seq=seq, name=name)
    assert o1.tm == pytest.approx(expected)





#!/usr/bin/env python3

"""\
Display a protocol for making the given fragment.

Usage:
    make <tag>... [--] [<options>...]

Arguments:
    <tag>
        The name of a plasmid or fragment in the database, e.g. p01 or f01.

    <options>
        Options that will be passed directly to all protocols invoked.

The protocol is parsed from the "Construction" column of these tables.  The 
syntax for a protocol is:

    <method>: [<key>=<value> ]...

The following methods are currently understood:

PCR: Polymerase chain reaction
    template: name of template (required)
    primers: forward and reverse primers, separated by a comma (required)
    Ta: annealing temperature (default: derived from primers)
    tx: extension time (default: derived from product length)
    scale: volume of the reaction (default: 10 µL)

INV: Inverse PCR
    Same as for PCR.

RE: Restriction enzyme digest
    template: name of template (required)
    enzyme: name of enzyme, must be sold by NEB (required)

IVT: In vitro transcription
    template: name of DNA template (required)

GG: Golden gate assembly
    bb: backbone plasmid (required)
    ins: comma-separated list of inserts (required)
    enzyme: Type IIS enzyme (default: BsaI)

GIB: Gibson assembly
    bb: backbone plasmid (required)
    ins: comma-separated list of inserts (required)
"""

import sys, re, shlex
import stepwise
import autoprop
from itertools import groupby
from subprocess import run
from statistics import mean
from numbers import Real
from more_itertools import one
from inform import Inform, Error, warn
from po4.model import load_db
from po4.protocols import *
from po4.errors import UsageError, QueryError

@autoprop
class Make:

    def __init__(self, db, tags=None, options=None):
        self.db = db
        self.tags = tags or []
        self.options = options or []

    def get_protocol(self):
        self._protocol = stepwise.Protocol()

        factories = {
                PcrProtocol:        self._make_pcr_command,
                InversePcrProtocol: self._make_inverse_pcr_command,
                DigestProtocol:     self._make_digest_command,
                AnnealProtocol:     self._make_anneal_command,
                GoldenGateProtocol: self._make_golden_gate_command,
                GibsonProtocol:     self._make_gibson_command,
                LigateProtocol:     self._make_ligate_command,
                IvtProtocol:        self._make_ivt_command,
        }

        constructs = [self.db[x] for x in self.tags]
        by_protocol_type = lambda x: type(x.protocol)

        for key, group in groupby(constructs, key=by_protocol_type):
            group = list(group)
            make_command = factories.get(key, lambda x:
                    warn(f"{key.name!r} protocols are not yet supported"))
            make_command(group)

        if not self._protocol:
            raise UsageError("no protocols found")

        return self._protocol

    def _make_pcr_command(self, constructs, cmd='pcr'):
        protocols = [x.protocol for x in constructs]

        def get_master_mix_flag():
            all_reagents = {'dna', 'primers'}
            master_mix = set(all_reagents)
            num_templates = len({x.template_tag for x in protocols})
            num_primers = len({x.primer_tags for x in protocols})

            if num_templates > 1:
                master_mix.discard('dna')
            if num_primers > 1:
                master_mix.discard('primers')

            if master_mix == all_reagents:
                return []
            if not master_mix:
                return ['-M']

            return ['-m', ','.join(master_mix)]

        self._add_command(constructs, [
                cmd,
                join(x.template_tag for x in protocols),
                join(x.primer_tags[0] for x in protocols),
                join(x.primer_tags[1] for x in protocols),
                '-a', join(str_sig(x.annealing_temp_C) for x in protocols),
                '-x', str(max(int(x.extension_time_s) for x in protocols)),
                *get_num_rxns_flag(constructs),
                *get_master_mix_flag(),
                *get_volume_flag(constructs),
        ])

    def _make_inverse_pcr_command(self, constructs):
        self._make_pcr_command(constructs, cmd='invpcr')

    def _make_ivt_command(self, constructs):
        protocols = [x.protocol for x in constructs]
        
        def get_short_flag():
            transcript_lens = (len(x.product_seq) for x in protocols)
            return ['-s'] if all(x <= 300 for x in transcript_lens) else []

        return self._add_command(constructs, [
                'ivt',
                *(x.template_tag for x in protocols),
                *get_short_flag(),
        ])

    def _make_digest_command(self, constructs):
        for enz, group in groupby(constructs, key=lambda x: x.protocol.enzymes):
            group = list(group)
            self._add_command(group, [
                    'digest',
                    join(x.protocol.template_tag for x in group),
                    join(enz),
            ])

    def _make_anneal_command(self, constructs):
        protocols = [x.protocol for x in constructs]

        def get_conc_flag():
            return get_unanimous_flag(
                    '-c', constructs, lambda x: x.conc_uM,
                    "reactions have different oligo concentrations",
            )

        def get_stock_conc_flag():
            return get_unanimous_flag(
                    '-C', constructs, lambda x: x.stock_uM,
                    "reactions have different oligo stock concentrations",
            )

        def get_master_mix_flag():
            if len(protocols) == 1:
                return []

            master_mix = []
            for i in range(2):
                unique_oligos = set(x.oligo_tags[i] for x in protocols)
                if len(unique_oligos) == 1:
                    master_mix.append(str(i+1))

            return ['-m', ','.join(master_mix)] if master_mix else []

        self._add_command(constructs, [
                'anneal',
                join(x.oligo_tags[0] for x in protocols),
                join(x.oligo_tags[1] for x in protocols),
                *get_num_rxns_flag(constructs),
                *get_volume_flag(constructs),
                *get_conc_flag(),
                *get_stock_conc_flag(),
                *get_master_mix_flag(),
        ])

    def _make_assembly_command(self, constructs, cmd, flags=()):
        protocols = [x.protocol for x in constructs]

        def fragment_from_tags(tags):
            def get_conc_str(tag):
                try:
                    conc = self.db[tag].conc_str
                    return re.sub(r'\s*ng/[µu]L$', '', conc)
                except QueryError:
                    return '50'

            def mean_if_similar(x, too_diff):
                if min(x) < 0.5 * max(x): raise too_diff
                return mean(x)

            def tabulate(x):
                from textwrap import indent
                from tabulate import tabulate
                return indent(
                        tabulate(x.items(), tablefmt='plain'),
                        '    ',
                )

            name = join(tags)

            concs = {
                    x: get_conc_str(x)
                    for x in tags
            }
            conc = one(
                    set(concs.values()),
                    too_long=UsageError(f"fragments have different concentrations:\n{tabulate(concs)}"),
            )

            # If the concentration happens to be in moles, we don't need to 
            # worry about the lengths of the fragments.  A bit hacky.
            if conc.endswith('M'):
                return f'{name}:{conc}'

            lengths = {
                    x: self.db[x].length
                    for x in tags
            }
            length = mean_if_similar(
                    lengths.values(),
                    too_diff=UsageError(f"fragments have concentrations in ng/µL and lengths that differ by more than 50%:\n{tabulate(lengths)}"),
            )

            return f'{name}:{conc}:{length}'

        def fragment_names():
            from itertools import count
            yield 'bb'
            for i in count(1):
                yield str(i)

        num_inserts = one(
                {len(x.insert_tags) for x in protocols},
                too_long=UsageError("all assemblies must have the same number of inserts"),
        )
        tag_groups = [
                [x.backbone_tag for x in protocols]
        ] + [
                [x.insert_tags[i] for x in protocols]
                for i in range(num_inserts)
        ]
        fragments = [
                fragment_from_tags(x)
                for x in tag_groups
        ]

        def get_master_mix_flag():
            master_mix = []
            for name, tags in zip(fragment_names(), tag_groups):
                if len(set(tags)) == 1:
                    master_mix.append(name)

            if master_mix and len(protocols) > 1:
                return ['-m', ','.join(master_mix)]
            else:
                return []

        stepwise_cmd = [
                cmd,
                *fragments,
                *get_num_rxns_flag(constructs),
                *get_volume_flag(constructs),
                *get_master_mix_flag(),
                *flags
        ]

        self._add_command(constructs, stepwise_cmd)

    def _make_golden_gate_command(self, constructs):
        self._make_assembly_command(
                constructs,
                cmd='golden_gate',
                flags=[
                    *get_unanimous_flag('-e', constructs, lambda x: x.enzyme, "assemblies use different enzymes"),
                ],
        )

    def _make_gibson_command(self, constructs):
        self._make_assembly_command(
                constructs,
                cmd='gibson',
        )

    def _make_ligate_command(self, constructs):
        protocols = [x.protocol for x in constructs]

        def get_kinase_flag():
            use_kinase = any(x.use_kinase for x in protocols)
            return ['-k'] if use_kinase else []

        self._make_assembly_command(
                constructs,
                cmd='ligate',
                flags=[
                    *get_kinase_flag(),
                ],
        )

    def _add_command(self, constructs, cmd):
        assert constructs
        assert cmd

        self._protocol += stepwise.load([*cmd, *self.options])

        label = f"Label the {plural(constructs):product/s}:"
        tags = ','.join(str(x.tag) for x in constructs)
        step = f'{label} {tags}'

        self._protocol += step if len(step) < 49 else f'{label}\n{tags}'

def join(items):
    items = list(items)
    if len(set(items)) == 1:
        return items[0]
    else:
        return ','.join(items)

def str_sig(value):
    return str(float(value)).strip('0').rstrip('.')


def get_volume_flag(constructs):
    return get_unanimous_flag(
            '-v',
            constructs,
            lambda x: x.volume_uL,
            "reactions have different volumes"
    )

def get_num_rxns_flag(constructs, default=1):
    n = len(constructs)
    return ['-n', str(n)] if n != default else []

def get_unanimous_flag(flag, constructs, value_getter, error):
    value = get_unanimous_value(constructs, value_getter, error)
    value_str = str_sig(value) if isinstance(value, Real) else str(value)
    return [flag, value_str] if value else []

def get_unanimous_value(constructs, value_getter, error):
    return one(
            {value_getter(x.protocol) for x in constructs},
            too_long=UsageError(f"{error}: {','.join(repr(x.tag) for x in constructs)}"),
    )


if __name__ == '__main__':
    import docopt

    try: 
        i = sys.argv.index('--')
        args = docopt.docopt(__doc__, sys.argv[1:i])
        args['<options>'] = sys.argv[i:][1:]
    except ValueError:
        args = docopt.docopt(__doc__)
        args['<options>'] = []

    db = load_db()
    make = Make(db)
    make.tags = args['<tag>']
    make.options = args['<options>']

    try:
        print(make.protocol)
    except Error as err:
        err.report()



#!/usr/bin/env python3

import functools
from inform import Error, get_culprit

class Po4Error(Error):

    def __init__(self, *args, **kwargs):
        if 'culprit' not in kwargs:
            kwargs['culprit'] = get_culprit()
        super().__init__(*args, **kwargs)

class LoadError(Po4Error):
    pass

class QueryError(Po4Error):
    pass

class ParseError(QueryError):
    pass

class CheckError(Po4Error):
    pass

class UsageError(Po4Error):
    pass



class only_raise:
    """
    Guarantee that the decorated function can only raise the given type of 
    exception.

    Any unhandled exception raised by the decorated function will be caught and 
    re-raised using an exception of the given type. 
    """

    def __init__(self, err_cls):
        self.err_cls = err_cls

    def __call__(self, f):

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except self.err_cls as err:
                raise err from None
            except Exception as err:
                raise self.err_cls(str(err)) from err

        return wrapper



#!/usr/bin/env python3

import math
import autoprop
import inform
from inform import plural
from statistics import mean
from more_itertools import one
from .model import Plasmid, Fragment, Oligo
from .errors import ParseError, QueryError
from .utils import *

@autoprop
class Protocol:
    is_product_double_stranded = True
    is_product_phosphorylated = True
    subclasses = {}

    def __init__(self, db):
        self._db = db

    def __init_subclass__(cls):
        if x := getattr(cls, 'name', None):
            Protocol.subclasses[x] = cls

    @classmethod
    def from_text(cls, db, protocol_str):
        method, params_str = protocol_str.split(':', 1)
        params = parse_params(params_str)

        try:
            subcls = cls.subclasses[method]
        except KeyError:
            raise ParseError("unknown protocol type", culprit=method)

        return subcls.from_params(db, params)

    def get_db(self):
        # Allow `self._db` to be a callable that returns a database.  This 
        # allows `Construct.__init__()` to instantiate protocols before it has 
        # access to a database.
        return self._db() if callable(self._db) else self._db

    def get_product_seq(self):
        raise NotImplementedError(self.__class__)


@autoprop
class PcrProtocol(Protocol):
    name = "PCR"

    def __init__(self, db, template, primers, Ta=None, tx=None, volume=None):
        super().__init__(db)
        self._template_tag = template
        self._primer_tags = primers
        self._annealing_temp_C = Ta
        self._extension_time_s = tx
        self._volume_uL = volume

    @classmethod
    def from_params(cls, db, params):
        pf = get_tag_pattern(Plasmid, Fragment)
        o = get_tag_pattern(Oligo)

        with inform.add_culprit(cls.name):
            pcr = cls(
                    db,
                    parse_param(params, 'template', pf),
                    parse_param(params, 'primers', fr'({o}),\s*({o})'),
            )

            if 'Ta' in params:
                pcr._annealing_temp_C = parse_temp_C(params['Ta'])
            
            if 'tx' in params:
                pcr._extension_time_s = parse_time_s(params['tx'])

            if 'volume' in params:
                pcr._volume_uL = parse_volume_uL(params['volume'])

        return pcr
        
    def get_template(self):
        return self.db[self.template_tag]

    def get_template_tag(self):
        return self._template_tag

    def get_template_seq(self):
        return self.template.seq

    def get_primers(self):
        return [self.db[x] for x in self.primer_tags]

    def get_primer_tags(self):
        return self._primer_tags

    def get_primer_seqs(self):
        return [x.seq for x in self.primers]

    def get_product_seq(self):
        seq = self.template_seq
        primers = p = self.primer_seqs
        primer_pairs = [
                (p[0], p[1].reverse_complement()),
                (p[1], p[0].reverse_complement()),
        ]

        for fwd, rev in primer_pairs:
            # Assume perfect complementarity in the last 15 bases.  This is a 
            # bit of a hack...
            i = seq.find(fwd[-15:  ])
            j = seq.find(rev[   :15])

            if i > j and self.template.is_linear:
                continue
            if i >= 0 and j >= 0:
                break
        else:
            raise QueryError(f"{self.primer_tags[0]!r} and {self.primer_tags[1]!r} do not amplify {self.template_tag!r}")

        if i < j:
            return fwd[:-15] + seq[i:j] + rev
        else:
            return fwd[:-15] + seq[i:] + seq[:j] + rev

    def get_product_len(self):
        return len(self.product_seq)

    def get_annealing_temp_C(self):
        if self._annealing_temp_C:
            return self._annealing_temp_C

        tms = [x.tm for x in self.primers]
        return min(tms) + 1

    def get_extension_time_s(self):
        if self._extension_time_s:
            return self._extension_time_s

        time_sec = 30 * self.product_len / 1000
        if time_sec <= 10: return 10
        if time_sec <= 15: return 15
        return 30 * math.ceil(time_sec / 30)

    def get_volume_uL(self):
        return self._volume_uL


@autoprop
class InversePcrProtocol(PcrProtocol):
    name = "INV"

@autoprop
class DigestProtocol(Protocol):
    name = "RE"

    def __init__(self, db, template, enzymes, size=None):
        super().__init__(db)
        self._template_tag = template
        self._enzymes = enzymes
        self._product_size = size

    @classmethod
    def from_params(cls, db, params):
        p = get_tag_pattern(Plasmid)

        with inform.add_culprit(cls.name):
            self = cls(
                    db,
                    parse_param(params, 'template', p),
                    parse_param(params, 'enzymes', r'[\w\d,-]+').split(','),
            )
            if 'size' in params:
                self._product_size = parse_size_bp(params['size'])

            return self

    def get_template(self):
        return self.db[self._template_tag]

    def get_template_tag(self):
        return self._template_tag

    def get_template_seq(self):
        return self.template.seq

    def get_enzymes(self):
        return self._enzymes

    def get_product_seqs(self):
        from more_itertools import pairwise, flatten
        from Bio.Restriction import RestrictionBatch

        if not self.enzymes:
            raise QueryError("no enzymes specified.", culprit=self.name)

        enzymes_str = ','.join(repr(x) for x in self.enzymes)
        enzymes = [
                re.sub('-HF(v2)?$', '', x)
                for x in self.enzymes
        ]

        try:
            batch = RestrictionBatch(enzymes)
        except ValueError:
            raise QueryError(f"unknown enzyme(s): {enzymes_str}", culprit=self.name) from None

        seq = self.template_seq
        sites = [x-1 for x in flatten(batch.search(seq).values())]

        if not sites:
            raise QueryError(f"{enzymes_str} {plural(enzymes):/does/do} not cut {self.template_tag!r}", culprit=self.name)

        sites += [0, len(seq)] if self.template.is_linear else []
        sites = sorted(sites)

        if len(sites) == 1:
            site, = sites

        seqs = []
        for i,j in pairwise(sorted(sites)):
            seqs.append(seq[i:j])

        if self.template.is_circular:
            wrap_around = seq[sites[-1]:] + seq[:sites[0]]
            seqs.append(wrap_around)

        return seqs

    def get_product_seq(self):
        target_size = self._product_size or len(self.template_seq)
        return min(self.product_seqs, key=lambda x: abs(target_size - len(x)))


@autoprop
class AnnealProtocol(Protocol):
    name = "ANNEAL"

    def __init__(self, db, oligos, volume=None, conc=None, stock=None):
        super().__init__(db)
        self._oligo_tags = oligos
        self._volume_uL = volume
        self._conc_uM = conc
        self._stock_uM = stock

    @classmethod
    def from_params(cls, db, params):
        o = get_tag_pattern(Oligo)

        with inform.add_culprit(cls.name):
            anneal = cls(
                    db,
                    parse_param(params, 'oligos', fr'({o}),\s*({o})'),
            )
            mw = mean(x.mw for x in anneal.oligos)

            if 'volume' in params:
                anneal._volume_uL = parse_volume_uL(params['volume'])
            if 'conc' in params:
                anneal._conc_uM = parse_conc_uM(params['conc'], mw)
            if 'stock' in params:
                anneal._stock_uM = parse_conc_uM(params['stock'], mw)

        return anneal
        
    def get_product_seq(self):
        # This assumes that the two oligos are perfectly complementary, which 
        # may not be the case (e.g. there might be overhangs on either side).  
        # But there's no way to represent overhangs in PO₄ anyways, and this 
        # approximation should be close enough for most purposes.
        return self.oligo_seqs[0]

    def get_oligos(self):
        return [self.db[x] for x in self.oligo_tags]

    def get_oligo_tags(self):
        return self._oligo_tags

    def get_oligo_seqs(self):
        return [x.seq for x in self.oligos]

    def get_volume_uL(self):
        return self._volume_uL

    def get_conc_uM(self):
        return self._conc_uM

    def get_stock_uM(self):
        return self._stock_uM


@autoprop
class AssemblyProtocol(Protocol):
    allowed_tags = get_tag_pattern(Plasmid, Fragment)

    def __init__(self, db, backbone, inserts, volume=None):
        super().__init__(db)
        self._backbone_tag = backbone
        self._insert_tags = inserts
        self._volume_uL = volume

    @classmethod
    def from_params(cls, db, params):
        x = cls.allowed_tags

        with inform.add_culprit(cls.name):
            self = cls(
                    db,
                    parse_param(params, 'bb', x),
                    parse_param(params, 'ins', fr'{x}(?:,{x})*').split(','),
            )
            if 'volume' in params:
                self._volume_uL = parse_volume_uL(params['volume'])

            return self

    def get_backbone(self):
        return self.db[self.backbone_tag]

    def get_backbone_tag(self):
        return self._backbone_tag

    def get_backbone_seq(self):
        return self.backbone.seq

    def get_inserts(self):
        return [self.db[x] for x in self._insert_tags]

    def get_insert_tags(self):
        return self._insert_tags

    def get_insert_seqs(self):
        return [x.seq for x in self.inserts]

    def get_volume_uL(self):
        return self._volume_uL


@autoprop
class GoldenGateProtocol(AssemblyProtocol):
    name = "GG"

    def __init__(self, db, backbone, inserts, volume=None, enzyme=None):
        super().__init__(db, backbone, inserts, volume)
        self._enzyme = enzyme

    @classmethod
    def from_params(cls, db, params):
        gg = super().from_params(db, params)

        with inform.add_culprit(cls.name):
            if 'enzyme' in params:
                gg._enzyme = parse_param(params, 'enzyme', r'[\w\d-]+')

        return gg

    def get_enzyme(self):
        return self._enzyme or None

@autoprop
class GibsonProtocol(AssemblyProtocol):
    name = "GIB"


@autoprop
class LigateProtocol(AssemblyProtocol):
    name = "LIG"
    allowed_tags = get_tag_pattern(Fragment)

    def __init__(self, db, backbone, inserts, volume=None, use_kinase=False):
        super().__init__(db, backbone, inserts, volume)
        self._use_kinase = use_kinase

    @classmethod
    def from_params(cls, db, params):
        kld = super().from_params(db, params)

        with inform.add_culprit(cls.name):
            if 'kinase' in params:
                kld._use_kinase = parse_bool(params['kinase'])

        return kld
        
    def get_use_kinase(self):
        return self._use_kinase


@autoprop
class IvtProtocol(Protocol):
    name = "IVT"
    is_product_double_stranded = False

    # Caveats:
    # - Only recognizes the T7 promoter.
    # - Doesn't recognize any terminators; always reads to the end.
    # - Only works with fragments (see above).

    def __init__(self, db, template):
        super().__init__(db)
        self._template_tag = template

    @classmethod
    def from_params(cls, db, params):
        f = get_tag_pattern(Fragment)

        with inform.add_culprit(cls.name):
            return cls(
                    db,
                    parse_param(params, 'template', f),
            )

    def get_template(self):
        return self.db[self.template_tag]

    def get_template_tag(self):
        return self._template_tag

    def get_template_seq(self):
        return self.template.seq

    def get_product_seq(self):
        t7_promoter = 'TAATACGACTCACTATA'
        seq = str(self.template_seq)
        i = seq.find(t7_promoter)
        j = i + len(t7_promoter)

        if i < 0:
            raise QueryError(f"{self.template_tag!r} does not contain a T7 promoter ({t7_promoter!r}).", culprit=self.name)

        return DnaSeq(seq[j:]).transcribe()


@autoprop
class IdtProtocol(Protocol):
    name = "IDT"

    def __init__(self, db, seq):
        super().__init__(db)
        self._product_seq = seq

    @classmethod
    def from_params(cls, db, params):
        pfo = get_tag_pattern()

        with inform.add_culprit(cls.name):
            return cls(
                    db,
                    parse_param(params, 'seq', '[ATCGUatcgu]+', None),
            )

    def get_product_seq(self):
        return self._product_seq



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['po4', 'po4.cli', 'po4.loaders', 'po4.stepwise']

package_data = \
{'': ['*']}

install_requires = \
['appdirs',
 'autoprop',
 'autosnapgene',
 'Biopython',
 'configurator',
 'docopt',
 'inform>=1.21',
 'more_itertools>=8.0',
 'pandas',
 'setuptools',
 'stepwise',
 'stepwise_mol_bio',
 'tabulate',
 'toml',
 'tqdm',
 'voluptuous',
 'xlrd>=1.0']

extras_require = \
{'docs': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'tests': ['pytest', 'pytest-cov', 'coveralls']}

entry_points = \
{'console_scripts': ['po4 = po4.cli.main:main'],
 'po4.databases': ['excel = po4.loaders.excel'],
 'stepwise.protocols': ['protocols = po4.stepwise:Plugin']}

setup(name='po4',
      version='0.2.0',
      description='Database of DNA sequences.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/po4',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.8',
     )



#!/usr/bin/env python3

import re
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna
from more_itertools import split_when
from .errors import ParseError, only_raise

DnaSeq = lambda x: Seq(normalize_seq(x), generic_dna)
no_default = object()

def normalize_seq(raw_seq):
    # Ignore nonstandard nucleotides; they're too hard to deal with properly.

    # This regular expression works because `re.sub()` only substitutes the 
    # left-most occurrence of any overlapping patterns.  The non-greedy * is 
    # necessary to avoid eliding everything between the first and last 
    # nonstandard nucleotide.
    seq = re.sub(r'/.*?/', 'X', raw_seq)

    return seq.upper()

def get_tag_prefixes(*args):
    from .model import Construct

    def add(prefixes, cls):
        if cls.tag_prefix:
            prefixes.add(cls.tag_prefix)
        for subcls in cls.__subclasses__():
            add(prefixes, subcls)

    prefixes = set()
    for cls in args or [Construct]:
        add(prefixes, cls)
    
    return prefixes

def get_tag_pattern(*args):
    prefix_chars = ''.join(sorted(get_tag_prefixes(*args)))
    return fr'[{prefix_chars}]\d+'

@only_raise(ParseError)
def parse_tag(tag_str):
    from .model import Tag, get_tag_prefixes

    if isinstance(tag_str, Tag):
        return tag_str
    if isinstance(tag_str, tuple):
        return Tag(*tag_str)

    pfo = get_tag_prefixes()
    if m := re.fullmatch(fr'\s*(?P<type>[{pfo}])(?P<id>\d+)\s*', tag_str):
        return Tag(m.group('type'), int(m.group('id')))
    else:
        raise ParseError(f"expected a tag (e.g. 'p100'), not {tag_str!r}")

@only_raise(ParseError)
def parse_params(params_str, *, sep=r'\s', on_duplicate=None, on_unmatched=None):
    params = {}
    param_pattern = fr'''\s*
            (?P<key>\w+)=(
              (?P<value>[^"'][^{sep}]*)|
              (?P<quote>["'])(?P<value_quoted>.*?)(?P=quote)
            )
            ({sep}|$)
    '''

    unmatched = set(range(len(params_str)))

    for m in re.finditer(param_pattern, params_str, re.VERBOSE):
        key = m.group('key')
        value = m.group('value') or m.group('value_quoted')

        if key in params:
            if on_duplicate:
                on_duplicate(params, key, value)
            else:
                raise ParseError(f"duplicate key {key!r} in {params_str!r}")

        params[key] = value
        unmatched -= set(range(m.start(), m.end()))

    if unmatched:
        if on_unmatched:
            on_unmatched(params_str, unmatched)
        else:
            groups = split_when(
                    sorted(unmatched), 
                    lambda a, b: b != a + 1
            )
            substrs = ', '.join(
                    repr(params_str[x[0]:x[-1]+1])
                    for x in groups
            )
            if substrs == repr(params_str):
                raise ParseError(f"can't parse {params_str!r}")
            else:
                raise ParseError(f"can't parse {params_str!r}: didn't expect {substrs}")

    return params

def parse_param(params, key, pattern, default=no_default):
    if key not in params:
        if default is not no_default:
            return default
        else:
            raise ParseError(f"missing required {key!r} parameter")

    if m := re.fullmatch(pattern, p := params[key]):
        return m.groups() or m.group()
    else:
        raise ParseError(f"expected {key}≈{pattern!r}, found {p!r}")

@only_raise(ParseError)
def parse_bool(bool_str):
    if bool_str.lower() in ('1', 'y', 'yes', 'true'):
        return True
    if bool_str.lower() in ('0', 'n', 'no', 'false'):
        return False

    raise ParseError(f"can't interpret {bool_str!r} as a bool")

@only_raise(ParseError)
def parse_time_s(time_str):
    time_units = {
            's':        1,
            'sec':      1,
            'second':   1,
            'seconds':  1,
            'm':        60,
            'min':      60,
            'minute':   60,
            'minutes':  60,
            'h':        60*60,
            'hr':       60*60,
            'hour':     60*60,
            'hours':    60*60,
    }
    time_pattern_1 = fr'(?P<time>\d+)\s*(?P<unit>{"|".join(time_units)})'
    time_pattern_2 = fr'(?P<min>\d+)m(?P<sec>\d+)'

    if m := re.fullmatch(time_pattern_1, time_str):
        return int(m.group('time')) * time_units[m.group('unit')]
    if m := re.fullmatch(time_pattern_2, time_str):
        return 60 * int(m.group('min')) + int(m.group('sec'))

    raise ParseError(f"can't interpret {time_str!r} as a time, did you forget a unit?")

@only_raise(ParseError)
def parse_temp_C(temp_str):
    temp_pattern = fr'(?P<temp>\d+)\s*°?C'

    if m := re.fullmatch(temp_pattern, temp_str):
        return int(m.group('temp'))

    raise ParseError(f"can't interpret {temp_str!r} as a temperature, did you forget a unit?")

@only_raise(ParseError)
def parse_volume_uL(vol_str):
    vol_pattern = fr'(?P<vol>\d+)\s*(?P<si_prefix>[nµum])L'
    si_prefixes = {
            'n': 1e-3,
            'u': 1,
            'µ': 1,
            'm': 1e3,
    }

    if m := re.fullmatch(vol_pattern, vol_str):
        return float(m.group('vol')) * si_prefixes[m.group('si_prefix')]

    raise ParseError(f"can't interpret {vol_str!r} as a volume, did you forget a unit?")

@only_raise(ParseError)
def parse_conc_nM(conc_str, mw):
    conc_pattern = r'(?P<conc>\d+)\s?(?P<unit>[nuµ]M|ng/[uµ]L)'
    unit_conversion = {
            'nM': 1,
            'uM': 1e3,
            'µM': 1e3,
            'ng/uL': 1e6 / mw,
            'ng/µL': 1e6 / mw,
    }

    if m := re.match(conc_pattern, conc_str):
        return float(m.group('conc')) * unit_conversion[m.group('unit')]
    else:
        raise ParseError(f"can't interpret {conc_str!r} as a concentration, did you forget a unit?")

@only_raise(ParseError)
def parse_conc_uM(conc_str, mw):
    return parse_conc_nM(conc_str, mw) / 1000

@only_raise(ParseError)
def parse_conc_ng_uL(conc_str, mw):
    conc_pattern = r'(?P<conc>\d+)\s?(?P<unit>[nuµ]M|ng/[uµ]L)'
    unit_conversion = {
            'ng/uL': 1,
            'ng/µL': 1,
            'nM': mw / 1e6,
            'uM': mw / 1e3,
            'µM': mw / 1e3,
    }

    if m := re.match(conc_pattern, conc_str):
        return float(m.group('conc')) * unit_conversion[m.group('unit')]
    else:
        raise ParseError(f"can't interpret {conc_str!r} as a concentration, did you forget a unit?")

@only_raise(ParseError)
def parse_size_bp(size_str):
    bp_parsers = [
            (
                fr'(?P<size>\d+)\s*bp',
                int,
            ), (
                fr'(?P<size>\d+.?\d*)\s*kb',
                lambda x: int(float(x) * 1000),
            ),
    ]

    for pattern, size_from_str in bp_parsers:
        if m := re.fullmatch(pattern, size_str):
            return size_from_str(m.group('size'))

    raise ParseError(f"can't interpret {size_str!r} as a size in base pairs, did you forget a unit?")




#!/usr/bin/env python3

import pytest
from pathlib import Path
from po4 import LoadError, load_db
from utils import parametrize_via_toml

@parametrize_via_toml('test_load.toml')
def test_load_db_err(config, error):
    with pytest.raises(LoadError, match=error):
        load_db(config=config)
